@@ -2,23 +2,24 @@
  * Copyright © 2011 Kristian Høgsberg
  * Copyright © 2011 Collabora, Ltd.
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
 #include "config.h"
@@ -39,24 +40,39 @@
 #include <libgen.h>
 #include <ctype.h>
 #include <time.h>
+#include <assert.h>
 
 #include <wayland-client.h>
 #include "window.h"
-#include "../shared/cairo-util.h"
-#include "../shared/config-parser.h"
+#include "shared/cairo-util.h"
+#include "shared/config-parser.h"
+#include "shared/helpers.h"
+#include "shared/xalloc.h"
+#include "shared/zalloc.h"
 
-#include "desktop-shell-client-protocol.h"
+#include "weston-desktop-shell-client-protocol.h"
+
+#define DEFAULT_CLOCK_FORMAT CLOCK_FORMAT_MINUTES
 
 extern char **environ; /* defined by libc */
 
+enum clock_format {
+	CLOCK_FORMAT_MINUTES,
+	CLOCK_FORMAT_SECONDS,
+	CLOCK_FORMAT_NONE
+};
+
 struct desktop {
 	struct display *display;
-	struct desktop_shell *shell;
-	uint32_t interface_version;
+	struct weston_desktop_shell *shell;
 	struct unlock_dialog *unlock_dialog;
 	struct task unlock_task;
 	struct wl_list outputs;
 
+	int want_panel;
+	enum weston_desktop_shell_panel_position panel_position;
+	enum clock_format clock_format;
+
 	struct window *grab_window;
 	struct widget *grab_widget;
 
@@ -70,7 +86,7 @@ struct desktop {
 
 struct surface {
 	void (*configure)(void *data,
-			  struct desktop_shell *desktop_shell,
+			  struct weston_desktop_shell *desktop_shell,
 			  uint32_t edges, struct window *window,
 			  int32_t width, int32_t height);
 };
@@ -82,6 +98,8 @@ struct panel {
 	struct wl_list launcher_list;
 	struct panel_clock *clock;
 	int painted;
+	enum weston_desktop_shell_panel_position panel_position;
+	enum clock_format clock_format;
 	uint32_t color;
 };
 
@@ -121,6 +139,8 @@ struct panel_clock {
 	struct panel *panel;
 	struct task clock_task;
 	int clock_fd;
+	char *format_string;
+	time_t refresh_timer;
 };
 
 struct unlock_dialog {
@@ -172,8 +192,7 @@ check_desktop_ready(struct window *window)
 	if (!desktop->painted && is_desktop_painted(desktop)) {
 		desktop->painted = 1;
 
-		if (desktop->interface_version >= 2)
-			desktop_shell_desktop_ready(desktop->shell);
+		weston_desktop_shell_desktop_ready(desktop->shell);
 	}
 }
 
@@ -356,7 +375,7 @@ panel_clock_redraw_handler(struct widget *widget, void *data)
 
 	time(&rawtime);
 	timeinfo = localtime(&rawtime);
-	strftime(string, sizeof string, "%a %b %d, %I:%M %p", timeinfo);
+	strftime(string, sizeof string, clock->format_string, timeinfo);
 
 	widget_get_allocation(widget, &allocation);
 	if (allocation.width == 0)
@@ -385,9 +404,9 @@ clock_timer_reset(struct panel_clock *clock)
 {
 	struct itimerspec its;
 
-	its.it_interval.tv_sec = 60;
+	its.it_interval.tv_sec = clock->refresh_timer;
 	its.it_interval.tv_nsec = 0;
-	its.it_value.tv_sec = 60;
+	its.it_value.tv_sec = clock->refresh_timer;
 	its.it_value.tv_nsec = 0;
 	if (timerfd_settime(clock->clock_fd, 0, &its, NULL) < 0) {
 		fprintf(stderr, "could not set timerfd\n: %m");
@@ -424,6 +443,19 @@ panel_add_clock(struct panel *panel)
 	panel->clock = clock;
 	clock->clock_fd = timerfd;
 
+	switch (panel->clock_format) {
+	case CLOCK_FORMAT_MINUTES:
+		clock->format_string = "%a %b %d, %I:%M %p";
+		clock->refresh_timer = 60;
+		break;
+	case CLOCK_FORMAT_SECONDS:
+		clock->format_string = "%a %b %d, %I:%M:%S %p";
+		clock->refresh_timer = 1;
+		break;
+	case CLOCK_FORMAT_NONE:
+		assert(!"not reached");
+	}
+
 	clock->clock_task.run = clock_func;
 	display_watch_fd(window_get_display(panel->window), clock->clock_fd,
 			 EPOLLIN, &clock->clock_task);
@@ -439,35 +471,81 @@ panel_resize_handler(struct widget *widget,
 {
 	struct panel_launcher *launcher;
 	struct panel *panel = data;
-	int x, y, w, h;
+	int bx = width / 2;
+	int by = height / 2;
+	int spacing = 10;
+	int x = spacing;
+	int y = spacing;
+	int w, h;
+	int horizontal = panel->panel_position == WESTON_DESKTOP_SHELL_PANEL_POSITION_TOP || panel->panel_position == WESTON_DESKTOP_SHELL_PANEL_POSITION_BOTTOM;
 
-	x = 10;
-	y = 16;
 	wl_list_for_each(launcher, &panel->launcher_list, link) {
 		w = cairo_image_surface_get_width(launcher->icon);
 		h = cairo_image_surface_get_height(launcher->icon);
+
+		if (horizontal)
+			y = by - h / 2;
+		else
+			x = bx - w / 2;
 		widget_set_allocation(launcher->widget,
-				      x, y - h / 2, w + 1, h + 1);
-		x += w + 10;
+				      x, y, w + 1, h + 1);
+		if (horizontal)
+			x += w + spacing;
+		else
+			y += h + spacing;
+	}
+
+	h = 20;
+
+	if (panel->clock_format == CLOCK_FORMAT_SECONDS)
+		w = 190;
+	else /* CLOCK_FORMAT_MINUTES */
+		w = 170;
+
+	if (horizontal) {
+		x = width - w - spacing;
+		y = by - h / 2;
+	} else {
+		x = bx - w / 2;
+		y = height - h - spacing;
 	}
-	h=20;
-	w=170;
 
 	if (panel->clock)
 		widget_set_allocation(panel->clock->widget,
-				      width - w - 8, y - h / 2, w + 1, h + 1);
+				      x, y, w + 1, h + 1);
 }
 
 static void
 panel_configure(void *data,
-		struct desktop_shell *desktop_shell,
+		struct weston_desktop_shell *desktop_shell,
 		uint32_t edges, struct window *window,
 		int32_t width, int32_t height)
 {
+	struct desktop *desktop = data;
 	struct surface *surface = window_get_user_data(window);
 	struct panel *panel = container_of(surface, struct panel, base);
 
-	window_schedule_resize(panel->window, width, 32);
+	switch (desktop->panel_position) {
+	case WESTON_DESKTOP_SHELL_PANEL_POSITION_TOP:
+	case WESTON_DESKTOP_SHELL_PANEL_POSITION_BOTTOM:
+		height = 32;
+		break;
+	case WESTON_DESKTOP_SHELL_PANEL_POSITION_LEFT:
+	case WESTON_DESKTOP_SHELL_PANEL_POSITION_RIGHT:
+		switch (desktop->clock_format) {
+		case CLOCK_FORMAT_NONE:
+			width = 32;
+			break;
+		case CLOCK_FORMAT_MINUTES:
+			width = 170;
+			break;
+		case CLOCK_FORMAT_SECONDS:
+			width = 190;
+			break;
+		}
+		break;
+	}
+	window_schedule_resize(panel->window, width, height);
 }
 
 static void
@@ -492,7 +570,8 @@ panel_destroy(struct panel *panel)
 	struct panel_launcher *tmp;
 	struct panel_launcher *launcher;
 
-	panel_destroy_clock(panel->clock);
+	if (panel->clock)
+		panel_destroy_clock(panel->clock);
 
 	wl_list_for_each_safe(launcher, tmp, &panel->launcher_list, link)
 		panel_destroy_launcher(launcher);
@@ -508,6 +587,7 @@ panel_create(struct desktop *desktop)
 {
 	struct panel *panel;
 	struct weston_config_section *s;
+	char *clock_format_option = NULL;
 
 	panel = xzalloc(sizeof *panel);
 
@@ -522,11 +602,16 @@ panel_create(struct desktop *desktop)
 	widget_set_redraw_handler(panel->widget, panel_redraw_handler);
 	widget_set_resize_handler(panel->widget, panel_resize_handler);
 
-	panel_add_clock(panel);
+	panel->panel_position = desktop->panel_position;
+	panel->clock_format = desktop->clock_format;
+	if (panel->clock_format != CLOCK_FORMAT_NONE)
+		panel_add_clock(panel);
+
+	free (clock_format_option);
 
 	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
-	weston_config_section_get_uint(s, "panel-color",
-				       &panel->color, 0xaa000000);
+	weston_config_section_get_color(s, "panel-color",
+					&panel->color, 0xaa000000);
 
 	panel_add_launchers(panel, desktop);
 
@@ -723,7 +808,7 @@ background_draw(struct widget *widget, void *data)
 
 static void
 background_configure(void *data,
-		     struct desktop_shell *desktop_shell,
+		     struct weston_desktop_shell *desktop_shell,
 		     uint32_t edges, struct window *window,
 		     int32_t width, int32_t height)
 {
@@ -858,6 +943,7 @@ unlock_dialog_create(struct desktop *desktop)
 {
 	struct display *display = desktop->display;
 	struct unlock_dialog *dialog;
+	struct wl_surface *surface;
 
 	dialog = xzalloc(sizeof *dialog);
 
@@ -882,8 +968,8 @@ unlock_dialog_create(struct desktop *desktop)
 	widget_set_touch_up_handler(dialog->button,
 				      unlock_dialog_touch_up_handler);
 
-	desktop_shell_set_lock_surface(desktop->shell,
-				       window_get_wl_surface(dialog->window));
+	surface = window_get_wl_surface(dialog->window);
+	weston_desktop_shell_set_lock_surface(desktop->shell, surface);
 
 	window_schedule_resize(dialog->window, 260, 230);
 
@@ -903,14 +989,14 @@ unlock_dialog_finish(struct task *task, uint32_t events)
 	struct desktop *desktop =
 		container_of(task, struct desktop, unlock_task);
 
-	desktop_shell_unlock(desktop->shell);
+	weston_desktop_shell_unlock(desktop->shell);
 	unlock_dialog_destroy(desktop->unlock_dialog);
 	desktop->unlock_dialog = NULL;
 }
 
 static void
 desktop_shell_configure(void *data,
-			struct desktop_shell *desktop_shell,
+			struct weston_desktop_shell *desktop_shell,
 			uint32_t edges,
 			struct wl_surface *surface,
 			int32_t width, int32_t height)
@@ -923,12 +1009,12 @@ desktop_shell_configure(void *data,
 
 static void
 desktop_shell_prepare_lock_surface(void *data,
-				   struct desktop_shell *desktop_shell)
+				   struct weston_desktop_shell *desktop_shell)
 {
 	struct desktop *desktop = data;
 
 	if (!desktop->locking) {
-		desktop_shell_unlock(desktop->shell);
+		weston_desktop_shell_unlock(desktop->shell);
 		return;
 	}
 
@@ -940,52 +1026,52 @@ desktop_shell_prepare_lock_surface(void *data,
 
 static void
 desktop_shell_grab_cursor(void *data,
-			  struct desktop_shell *desktop_shell,
+			  struct weston_desktop_shell *desktop_shell,
 			  uint32_t cursor)
 {
 	struct desktop *desktop = data;
 
 	switch (cursor) {
-	case DESKTOP_SHELL_CURSOR_NONE:
+	case WESTON_DESKTOP_SHELL_CURSOR_NONE:
 		desktop->grab_cursor = CURSOR_BLANK;
 		break;
-	case DESKTOP_SHELL_CURSOR_BUSY:
+	case WESTON_DESKTOP_SHELL_CURSOR_BUSY:
 		desktop->grab_cursor = CURSOR_WATCH;
 		break;
-	case DESKTOP_SHELL_CURSOR_MOVE:
+	case WESTON_DESKTOP_SHELL_CURSOR_MOVE:
 		desktop->grab_cursor = CURSOR_DRAGGING;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_TOP:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_TOP:
 		desktop->grab_cursor = CURSOR_TOP;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM:
 		desktop->grab_cursor = CURSOR_BOTTOM;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_LEFT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_LEFT:
 		desktop->grab_cursor = CURSOR_LEFT;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_RIGHT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_RIGHT:
 		desktop->grab_cursor = CURSOR_RIGHT;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_TOP_LEFT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_TOP_LEFT:
 		desktop->grab_cursor = CURSOR_TOP_LEFT;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_TOP_RIGHT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_TOP_RIGHT:
 		desktop->grab_cursor = CURSOR_TOP_RIGHT;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM_LEFT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM_LEFT:
 		desktop->grab_cursor = CURSOR_BOTTOM_LEFT;
 		break;
-	case DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM_RIGHT:
+	case WESTON_DESKTOP_SHELL_CURSOR_RESIZE_BOTTOM_RIGHT:
 		desktop->grab_cursor = CURSOR_BOTTOM_RIGHT;
 		break;
-	case DESKTOP_SHELL_CURSOR_ARROW:
+	case WESTON_DESKTOP_SHELL_CURSOR_ARROW:
 	default:
 		desktop->grab_cursor = CURSOR_LEFT_PTR;
 	}
 }
 
-static const struct desktop_shell_listener listener = {
+static const struct weston_desktop_shell_listener listener = {
 	desktop_shell_configure,
 	desktop_shell_prepare_lock_surface,
 	desktop_shell_grab_cursor
@@ -1015,14 +1101,12 @@ background_create(struct desktop *desktop)
 	window_set_user_data(background->window, background);
 	widget_set_redraw_handler(background->widget, background_draw);
 	widget_set_transparent(background->widget, 0);
-	window_set_preferred_format(background->window,
-				    WINDOW_PREFERRED_FORMAT_RGB565);
 
 	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
 	weston_config_section_get_string(s, "background-image",
 					 &background->image, NULL);
-	weston_config_section_get_uint(s, "background-color",
-				       &background->color, 0);
+	weston_config_section_get_color(s, "background-color",
+					&background->color, 0x00000000);
 
 	weston_config_section_get_string(s, "background-type",
 					 &type, "tile");
@@ -1073,7 +1157,7 @@ grab_surface_create(struct desktop *desktop)
 	window_set_user_data(desktop->grab_window, desktop);
 
 	s = window_get_wl_surface(desktop->grab_window);
-	desktop_shell_set_grab_surface(desktop->shell, s);
+	weston_desktop_shell_set_grab_surface(desktop->shell, s);
 
 	desktop->grab_widget =
 		window_add_widget(desktop->grab_window, desktop);
@@ -1160,41 +1244,22 @@ static const struct wl_output_listener output_listener = {
 	output_handle_scale
 };
 
-static int
-want_panel(struct desktop *desktop)
-{
-	struct weston_config_section *s;
-	char *location = NULL;
-	int ret = 1;
-
-	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
-	weston_config_section_get_string(s, "panel-location",
-					 &location, "top");
-
-	if (strcmp(location, "top") != 0)
-		ret = 0;
-
-	free(location);
-
-	return ret;
-}
-
 static void
 output_init(struct output *output, struct desktop *desktop)
 {
 	struct wl_surface *surface;
 
-	if (want_panel(desktop)) {
+	if (desktop->want_panel) {
 		output->panel = panel_create(desktop);
 		surface = window_get_wl_surface(output->panel->window);
-		desktop_shell_set_panel(desktop->shell,
-					output->output, surface);
+		weston_desktop_shell_set_panel(desktop->shell,
+					       output->output, surface);
 	}
 
 	output->background = background_create(desktop);
 	surface = window_get_wl_surface(output->background->window);
-	desktop_shell_set_background(desktop->shell,
-				     output->output, surface);
+	weston_desktop_shell_set_background(desktop->shell,
+					    output->output, surface);
 }
 
 static void
@@ -1202,7 +1267,7 @@ create_output(struct desktop *desktop, uint32_t id)
 {
 	struct output *output;
 
-	output = calloc(1, sizeof *output);
+	output = zalloc(sizeof *output);
 	if (!output)
 		return;
 
@@ -1226,12 +1291,14 @@ global_handler(struct display *display, uint32_t id,
 {
 	struct desktop *desktop = data;
 
-	if (!strcmp(interface, "desktop_shell")) {
-		desktop->interface_version = (version < 2) ? version : 2;
+	if (!strcmp(interface, "weston_desktop_shell")) {
 		desktop->shell = display_bind(desktop->display,
-					      id, &desktop_shell_interface,
-					      desktop->interface_version);
-		desktop_shell_add_listener(desktop->shell, &listener, desktop);
+					      id,
+					      &weston_desktop_shell_interface,
+					      1);
+		weston_desktop_shell_add_listener(desktop->shell,
+						  &listener,
+						  desktop);
 	} else if (!strcmp(interface, "wl_output")) {
 		create_output(desktop, id);
 	}
@@ -1290,6 +1357,48 @@ panel_add_launchers(struct panel *panel, struct desktop *desktop)
 	}
 }
 
+static void
+parse_panel_position(struct desktop *desktop, struct weston_config_section *s)
+{
+	char *position;
+
+	desktop->want_panel = 1;
+
+	weston_config_section_get_string(s, "panel-position", &position, "top");
+	if (strcmp(position, "top") == 0) {
+		desktop->panel_position = WESTON_DESKTOP_SHELL_PANEL_POSITION_TOP;
+	} else if (strcmp(position, "bottom") == 0) {
+		desktop->panel_position = WESTON_DESKTOP_SHELL_PANEL_POSITION_BOTTOM;
+	} else if (strcmp(position, "left") == 0) {
+		desktop->panel_position = WESTON_DESKTOP_SHELL_PANEL_POSITION_LEFT;
+	} else if (strcmp(position, "right") == 0) {
+		desktop->panel_position = WESTON_DESKTOP_SHELL_PANEL_POSITION_RIGHT;
+	} else {
+		/* 'none' is valid here */
+		if (strcmp(position, "none") != 0)
+			fprintf(stderr, "Wrong panel position: %s\n", position);
+		desktop->want_panel = 0;
+	}
+	free(position);
+}
+
+static void
+parse_clock_format(struct desktop *desktop, struct weston_config_section *s)
+{
+	char *clock_format;
+
+	weston_config_section_get_string(s, "clock-format", &clock_format, "");
+	if (strcmp(clock_format, "minutes") == 0)
+		desktop->clock_format = CLOCK_FORMAT_MINUTES;
+	else if (strcmp(clock_format, "seconds") == 0)
+		desktop->clock_format = CLOCK_FORMAT_SECONDS;
+	else if (strcmp(clock_format, "none") == 0)
+		desktop->clock_format = CLOCK_FORMAT_NONE;
+	else
+		desktop->clock_format = DEFAULT_CLOCK_FORMAT;
+	free(clock_format);
+}
+
 int main(int argc, char *argv[])
 {
 	struct desktop desktop = { 0 };
@@ -1304,6 +1413,8 @@ int main(int argc, char *argv[])
 	desktop.config = weston_config_parse(config_file);
 	s = weston_config_get_section(desktop.config, "shell", NULL, NULL);
 	weston_config_section_get_bool(s, "locking", &desktop.locking, 1);
+	parse_panel_position(&desktop, s);
+	parse_clock_format(&desktop, s);
 
 	desktop.display = display_create(&argc, argv);
 	if (desktop.display == NULL) {
@@ -1317,6 +1428,8 @@ int main(int argc, char *argv[])
 
 	/* Create panel and background for outputs processed before the shell
 	 * global interface was processed */
+	if (desktop.want_panel)
+		weston_desktop_shell_set_panel_position(desktop.shell, desktop.panel_position);
 	wl_list_for_each(output, &desktop.outputs, link)
 		if (!output->panel)
 			output_init(output, &desktop);
@@ -1332,7 +1445,7 @@ int main(int argc, char *argv[])
 	desktop_destroy_outputs(&desktop);
 	if (desktop.unlock_dialog)
 		unlock_dialog_destroy(desktop.unlock_dialog);
-	desktop_shell_destroy(desktop.shell);
+	weston_desktop_shell_destroy(desktop.shell);
 	display_destroy(desktop.display);
 
 	return 0;
